from latex_input.unicode_structs import CharacterFontVariant

cached_subscripts = {'i': 'ᵢ', 'r': 'ᵣ', 'u': 'ᵤ', 'v': 'ᵥ', 'β': 'ᵦ', 'γ': 'ᵧ', 'ρ': 'ᵨ', 'φ': 'ᵩ', 'χ': 'ᵪ', '0': '₀', '1': '₁', '2': '₂', '3': '₃', '4': '₄', '5': '₅', '6': '₆', '7': '₇', '8': '₈', '9': '₉', '+': '₊', '−': '₋', '=': '₌', '(': '₍', ')': '₎', 'a': 'ₐ', 'e': 'ₑ', 'o': 'ₒ', 'x': 'ₓ', 'ə': 'ₔ', 'h': 'ₕ', 'k': 'ₖ', 'l': 'ₗ', 'm': 'ₘ', 'n': 'ₙ', 'p': 'ₚ', 's': 'ₛ', 't': 'ₜ', 'j': 'ⱼ', 'а': '\U0001e051', 'б': '\U0001e052', 'в': '\U0001e053', 'г': '\U0001e054', 'д': '\U0001e055', 'е': '\U0001e056', 'ж': '\U0001e057', 'з': '\U0001e058', 'и': '\U0001e059', 'к': '\U0001e05a', 'л': '\U0001e05b', 'о': '\U0001e05c', 'п': '\U0001e05d', 'с': '\U0001e05e', 'у': '\U0001e05f', 'ф': '\U0001e060', 'х': '\U0001e061', 'ц': '\U0001e062', 'ч': '\U0001e063', 'ш': '\U0001e064', 'ъ': '\U0001e065', 'ы': '\U0001e066', 'ґ': '\U0001e067', 'і': '\U0001e068', 'ѕ': '\U0001e069', 'џ': '\U0001e06a'}

cached_superscripts = {'a': 'ᵃ', '2': '²', '3': '³', '1': '¹', 'o': 'ᵒ', 'h': 'ʰ', 'ɦ': 'ʱ', 'j': 'ʲ', 'r': 'ʳ', 'ɹ': 'ʴ', 'ɻ': 'ʵ', 'ʁ': 'ʶ', 'w': 'ʷ', 'y': 'ʸ', 'ɣ': 'ˠ', 'l': 'ˡ', 's': 'ˢ', 'x': 'ˣ', 'ʕ': 'ˤ', 'ნ': 'ჼ', 'A': 'ᴬ', 'Æ': 'ᴭ', 'B': 'ᴮ', 'D': 'ᴰ', 'E': 'ᴱ', 'Ǝ': 'ᴲ', 'G': 'ᴳ', 'H': 'ᴴ', 'I': 'ᴵ', 'J': 'ᴶ', 'K': 'ᴷ', 'L': 'ᴸ', 'M': 'ᴹ', 'N': 'ᴺ', 'O': 'ᴼ', 'Ȣ': 'ᴽ', 'P': 'ᴾ', 'R': 'ᴿ', 'T': 'ᵀ', 'U': 'ᵁ', 'W': 'ᵂ', 'ɐ': 'ᵄ', 'ɑ': 'ᵅ', 'ᴂ': 'ᵆ', 'b': 'ᵇ', 'd': 'ᵈ', 'e': 'ᵉ', 'ə': 'ᵊ', 'ɛ': 'ᵋ', 'ɜ': 'ᶟ', 'g': 'ᵍ', 'k': 'ᵏ', 'm': 'ᵐ', 'ŋ': 'ᵑ', 'ɔ': 'ᵓ', 'ᴖ': 'ᵔ', 'ᴗ': 'ᵕ', 'p': 'ᵖ', 't': 'ᵗ', 'u': 'ᵘ', 'ᴝ': 'ᵙ', 'ɯ': 'ᵚ', 'v': 'ᵛ', 'ᴥ': 'ᵜ', 'β': 'ᵝ', 'γ': 'ᵞ', 'δ': 'ᵟ', 'φ': 'ᵠ', 'χ': 'ᵡ', 'н': 'ᵸ', 'ɒ': 'ᶛ', 'c': 'ᶜ', 'ɕ': 'ᶝ', 'ð': 'ᶞ', 'f': 'ᶠ', 'ɟ': 'ᶡ', 'ɡ': 'ᶢ', 'ɥ': 'ᶣ', 'ɨ': 'ᶤ', 'ɩ': 'ᶥ', 'ɪ': 'ᶦ', 'ᵻ': 'ᶧ', 'ʝ': 'ᶨ', 'ɭ': 'ᶩ', 'ᶅ': 'ᶪ', 'ʟ': 'ᶫ', 'ɱ': 'ᶬ', 'ɰ': 'ᶭ', 'ɲ': 'ᶮ', 'ɳ': 'ᶯ', 'ɴ': 'ᶰ', 'ɵ': 'ᶱ', 'ɸ': 'ᶲ', 'ʂ': 'ᶳ', 'ʃ': 'ᶴ', 'ƫ': 'ᶵ', 'ʉ': 'ᶶ', 'ʊ': 'ᶷ', 'ᴜ': 'ᶸ', 'ʋ': 'ᶹ', 'ʌ': 'ᶺ', 'z': 'ᶻ', 'ʐ': 'ᶼ', 'ʑ': 'ᶽ', 'ʒ': 'ᶾ', 'θ': 'ᶿ', '0': '⁰', 'i': 'ⁱ', '4': '⁴', '5': '⁵', '6': '⁶', '7': '⁷', '8': '⁸', '9': '⁹', '+': '⁺', '−': '⁻', '=': '⁼', '(': '⁽', ')': '⁾', 'n': 'ⁿ', 'V': 'ⱽ', 'ⵡ': 'ⵯ', '一': '㆒', '二': '㆓', '三': '㆔', '四': '㆕', '上': '㆖', '中': '㆗', '下': '㆘', '甲': '㆙', '乙': '㆚', '丙': '㆛', '丁': '㆜', '天': '㆝', '地': '㆞', '人': '㆟', 'ъ': 'ꚜ', 'ь': 'ꚝ', 'ꝯ': 'ꝰ', 'C': '\ua7f2', 'F': '\ua7f3', 'Q': '\ua7f4', 'Ħ': 'ꟸ', 'œ': 'ꟹ', 'ꜧ': 'ꭜ', 'ꬷ': 'ꭝ', 'ɫ': 'ꭞ', 'ꭒ': 'ꭟ', 'ʍ': 'ꭩ', 'ː': '\U00010781', 'ˑ': '\U00010782', 'æ': '\U00010783', 'ʙ': '\U00010784', 'ɓ': '\U00010785', 'ʣ': '\U00010787', 'ꭦ': '\U00010788', 'ʥ': '\U00010789', 'ʤ': '\U0001078a', 'ɖ': '\U0001078b', 'ɗ': '\U0001078c', 'ᶑ': '\U0001078d', 'ɘ': '\U0001078e', 'ɞ': '\U0001078f', 'ʩ': '\U00010790', 'ɤ': '\U00010791', 'ɢ': '\U00010792', 'ɠ': '\U00010793', 'ʛ': '\U00010794', 'ħ': '\U00010795', 'ʜ': '\U00010796', 'ɧ': '\U00010797', 'ʄ': '\U00010798', 'ʪ': '\U00010799', 'ʫ': '\U0001079a', 'ɬ': '\U0001079b', '\U0001df04': '\U0001079c', 'ꞎ': '\U0001079d', 'ɮ': '\U0001079e', '\U0001df05': '\U0001079f', 'ʎ': '\U000107a0', '\U0001df06': '\U000107a1', 'ø': '\U000107a2', 'ɶ': '\U000107a3', 'ɷ': '\U000107a4', 'q': '\U000107a5', 'ɺ': '\U000107a6', '\U0001df08': '\U000107a7', 'ɽ': '\U000107a8', 'ɾ': '\U000107a9', 'ʀ': '\U000107aa', 'ʨ': '\U000107ab', 'ʦ': '\U000107ac', 'ꭧ': '\U000107ad', 'ʧ': '\U000107ae', 'ʈ': '\U000107af', 'ⱱ': '\U000107b0', 'ʏ': '\U000107b2', 'ʡ': '\U000107b3', 'ʢ': '\U000107b4', 'ʘ': '\U000107b5', 'ǀ': '\U000107b6', 'ǁ': '\U000107b7', 'ǂ': '\U000107b8', '\U0001df0a': '\U000107b9', '\U0001df1e': '\U000107ba', 'а': '\U0001e030', 'б': '\U0001e031', 'в': '\U0001e032', 'г': '\U0001e033', 'д': '\U0001e034', 'е': '\U0001e035', 'ж': '\U0001e036', 'з': '\U0001e037', 'и': '\U0001e038', 'к': '\U0001e039', 'л': '\U0001e03a', 'м': '\U0001e03b', 'о': '\U0001e03c', 'п': '\U0001e03d', 'р': '\U0001e03e', 'с': '\U0001e03f', 'т': '\U0001e040', 'у': '\U0001e041', 'ф': '\U0001e042', 'х': '\U0001e043', 'ц': '\U0001e044', 'ч': '\U0001e045', 'ш': '\U0001e046', 'ы': '\U0001e047', 'э': '\U0001e048', 'ю': '\U0001e049', 'ꚉ': '\U0001e04a', 'ә': '\U0001e04b', 'і': '\U0001e04c', 'ј': '\U0001e04d', 'ө': '\U0001e04e', 'ү': '\U0001e04f', 'ӏ': '\U0001e050', 'ҫ': '\U0001e06b', 'ꙑ': '\U0001e06c', 'ұ': '\U0001e06d'}

cached_character_font_variants = {'C': [CharacterFontVariant(text='ℂ', kind=2), CharacterFontVariant(text='ℭ', kind=4), CharacterFontVariant(text='𝐂', kind=17), CharacterFontVariant(text='𝐶', kind=24), CharacterFontVariant(text='𝑪', kind=25), CharacterFontVariant(text='𝒞', kind=144), CharacterFontVariant(text='𝓒', kind=145), CharacterFontVariant(text='𝕮', kind=21), CharacterFontVariant(text='𝖢', kind=80), CharacterFontVariant(text='𝗖', kind=81), CharacterFontVariant(text='𝘊', kind=88), CharacterFontVariant(text='𝘾', kind=89), CharacterFontVariant(text='𝙲', kind=48)], 'g': [CharacterFontVariant(text='ℊ', kind=128), CharacterFontVariant(text='𝐠', kind=17), CharacterFontVariant(text='𝑔', kind=24), CharacterFontVariant(text='𝒈', kind=25), CharacterFontVariant(text='𝓰', kind=145), CharacterFontVariant(text='𝔤', kind=20), CharacterFontVariant(text='𝕘', kind=18), CharacterFontVariant(text='𝖌', kind=21), CharacterFontVariant(text='𝗀', kind=80), CharacterFontVariant(text='𝗴', kind=81), CharacterFontVariant(text='𝘨', kind=88), CharacterFontVariant(text='𝙜', kind=89), CharacterFontVariant(text='𝚐', kind=48)], 'H': [CharacterFontVariant(text='ℋ', kind=128), CharacterFontVariant(text='ℌ', kind=4), CharacterFontVariant(text='ℍ', kind=2), CharacterFontVariant(text='𝐇', kind=17), CharacterFontVariant(text='𝐻', kind=24), CharacterFontVariant(text='𝑯', kind=25), CharacterFontVariant(text='𝓗', kind=145), CharacterFontVariant(text='𝕳', kind=21), CharacterFontVariant(text='𝖧', kind=80), CharacterFontVariant(text='𝗛', kind=81), CharacterFontVariant(text='𝘏', kind=88), CharacterFontVariant(text='𝙃', kind=89), CharacterFontVariant(text='𝙷', kind=48)], 'h': [CharacterFontVariant(text='ℎ', kind=0), CharacterFontVariant(text='𝐡', kind=17), CharacterFontVariant(text='𝒉', kind=25), CharacterFontVariant(text='𝒽', kind=144), CharacterFontVariant(text='𝓱', kind=145), CharacterFontVariant(text='𝔥', kind=20), CharacterFontVariant(text='𝕙', kind=18), CharacterFontVariant(text='𝖍', kind=21), CharacterFontVariant(text='𝗁', kind=80), CharacterFontVariant(text='𝗵', kind=81), CharacterFontVariant(text='𝘩', kind=88), CharacterFontVariant(text='𝙝', kind=89), CharacterFontVariant(text='𝚑', kind=48)], 'ħ': [CharacterFontVariant(text='ℏ', kind=0)], 'I': [CharacterFontVariant(text='ℐ', kind=128), CharacterFontVariant(text='ℑ', kind=4), CharacterFontVariant(text='𝐈', kind=17), CharacterFontVariant(text='𝐼', kind=24), CharacterFontVariant(text='𝑰', kind=25), CharacterFontVariant(text='𝓘', kind=145), CharacterFontVariant(text='𝕀', kind=18), CharacterFontVariant(text='𝕴', kind=21), CharacterFontVariant(text='𝖨', kind=80), CharacterFontVariant(text='𝗜', kind=81), CharacterFontVariant(text='𝘐', kind=88), CharacterFontVariant(text='𝙄', kind=89), CharacterFontVariant(text='𝙸', kind=48)], 'L': [CharacterFontVariant(text='ℒ', kind=128), CharacterFontVariant(text='𝐋', kind=17), CharacterFontVariant(text='𝐿', kind=24), CharacterFontVariant(text='𝑳', kind=25), CharacterFontVariant(text='𝓛', kind=145), CharacterFontVariant(text='𝔏', kind=20), CharacterFontVariant(text='𝕃', kind=18), CharacterFontVariant(text='𝕷', kind=21), CharacterFontVariant(text='𝖫', kind=80), CharacterFontVariant(text='𝗟', kind=81), CharacterFontVariant(text='𝘓', kind=88), CharacterFontVariant(text='𝙇', kind=89), CharacterFontVariant(text='𝙻', kind=48)], 'l': [CharacterFontVariant(text='ℓ', kind=128), CharacterFontVariant(text='𝐥', kind=17), CharacterFontVariant(text='𝑙', kind=24), CharacterFontVariant(text='𝒍', kind=25), CharacterFontVariant(text='𝓁', kind=144), CharacterFontVariant(text='𝓵', kind=145), CharacterFontVariant(text='𝔩', kind=20), CharacterFontVariant(text='𝕝', kind=18), CharacterFontVariant(text='𝖑', kind=21), CharacterFontVariant(text='𝗅', kind=80), CharacterFontVariant(text='𝗹', kind=81), CharacterFontVariant(text='𝘭', kind=88), CharacterFontVariant(text='𝙡', kind=89), CharacterFontVariant(text='𝚕', kind=48)], 'N': [CharacterFontVariant(text='ℕ', kind=2), CharacterFontVariant(text='𝐍', kind=17), CharacterFontVariant(text='𝑁', kind=24), CharacterFontVariant(text='𝑵', kind=25), CharacterFontVariant(text='𝒩', kind=144), CharacterFontVariant(text='𝓝', kind=145), CharacterFontVariant(text='𝔑', kind=20), CharacterFontVariant(text='𝕹', kind=21), CharacterFontVariant(text='𝖭', kind=80), CharacterFontVariant(text='𝗡', kind=81), CharacterFontVariant(text='𝘕', kind=88), CharacterFontVariant(text='𝙉', kind=89), CharacterFontVariant(text='𝙽', kind=48)], 'P': [CharacterFontVariant(text='ℙ', kind=2), CharacterFontVariant(text='𝐏', kind=17), CharacterFontVariant(text='𝑃', kind=24), CharacterFontVariant(text='𝑷', kind=25), CharacterFontVariant(text='𝒫', kind=144), CharacterFontVariant(text='𝓟', kind=145), CharacterFontVariant(text='𝔓', kind=20), CharacterFontVariant(text='𝕻', kind=21), CharacterFontVariant(text='𝖯', kind=80), CharacterFontVariant(text='𝗣', kind=81), CharacterFontVariant(text='𝘗', kind=88), CharacterFontVariant(text='𝙋', kind=89), CharacterFontVariant(text='𝙿', kind=48)], 'Q': [CharacterFontVariant(text='ℚ', kind=2), CharacterFontVariant(text='𝐐', kind=17), CharacterFontVariant(text='𝑄', kind=24), CharacterFontVariant(text='𝑸', kind=25), CharacterFontVariant(text='𝒬', kind=144), CharacterFontVariant(text='𝓠', kind=145), CharacterFontVariant(text='𝔔', kind=20), CharacterFontVariant(text='𝕼', kind=21), CharacterFontVariant(text='𝖰', kind=80), CharacterFontVariant(text='𝗤', kind=81), CharacterFontVariant(text='𝘘', kind=88), CharacterFontVariant(text='𝙌', kind=89), CharacterFontVariant(text='𝚀', kind=48)], 'R': [CharacterFontVariant(text='ℛ', kind=128), CharacterFontVariant(text='ℜ', kind=4), CharacterFontVariant(text='ℝ', kind=2), CharacterFontVariant(text='𝐑', kind=17), CharacterFontVariant(text='𝑅', kind=24), CharacterFontVariant(text='𝑹', kind=25), CharacterFontVariant(text='𝓡', kind=145), CharacterFontVariant(text='𝕽', kind=21), CharacterFontVariant(text='𝖱', kind=80), CharacterFontVariant(text='𝗥', kind=81), CharacterFontVariant(text='𝘙', kind=88), CharacterFontVariant(text='𝙍', kind=89), CharacterFontVariant(text='𝚁', kind=48)], 'Z': [CharacterFontVariant(text='ℤ', kind=2), CharacterFontVariant(text='ℨ', kind=4), CharacterFontVariant(text='𝐙', kind=17), CharacterFontVariant(text='𝑍', kind=24), CharacterFontVariant(text='𝒁', kind=25), CharacterFontVariant(text='𝒵', kind=144), CharacterFontVariant(text='𝓩', kind=145), CharacterFontVariant(text='𝖅', kind=21), CharacterFontVariant(text='𝖹', kind=80), CharacterFontVariant(text='𝗭', kind=81), CharacterFontVariant(text='𝘡', kind=88), CharacterFontVariant(text='𝙕', kind=89), CharacterFontVariant(text='𝚉', kind=48)], 'B': [CharacterFontVariant(text='ℬ', kind=128), CharacterFontVariant(text='𝐁', kind=17), CharacterFontVariant(text='𝐵', kind=24), CharacterFontVariant(text='𝑩', kind=25), CharacterFontVariant(text='𝓑', kind=145), CharacterFontVariant(text='𝔅', kind=20), CharacterFontVariant(text='𝔹', kind=18), CharacterFontVariant(text='𝕭', kind=21), CharacterFontVariant(text='𝖡', kind=80), CharacterFontVariant(text='𝗕', kind=81), CharacterFontVariant(text='𝘉', kind=88), CharacterFontVariant(text='𝘽', kind=89), CharacterFontVariant(text='𝙱', kind=48)], 'e': [CharacterFontVariant(text='ℯ', kind=128), CharacterFontVariant(text='ⅇ', kind=10), CharacterFontVariant(text='𝐞', kind=17), CharacterFontVariant(text='𝑒', kind=24), CharacterFontVariant(text='𝒆', kind=25), CharacterFontVariant(text='𝓮', kind=145), CharacterFontVariant(text='𝔢', kind=20), CharacterFontVariant(text='𝕖', kind=18), CharacterFontVariant(text='𝖊', kind=21), CharacterFontVariant(text='𝖾', kind=80), CharacterFontVariant(text='𝗲', kind=81), CharacterFontVariant(text='𝘦', kind=88), CharacterFontVariant(text='𝙚', kind=89), CharacterFontVariant(text='𝚎', kind=48)], 'E': [CharacterFontVariant(text='ℰ', kind=128), CharacterFontVariant(text='𝐄', kind=17), CharacterFontVariant(text='𝐸', kind=24), CharacterFontVariant(text='𝑬', kind=25), CharacterFontVariant(text='𝓔', kind=145), CharacterFontVariant(text='𝔈', kind=20), CharacterFontVariant(text='𝔼', kind=18), CharacterFontVariant(text='𝕰', kind=21), CharacterFontVariant(text='𝖤', kind=80), CharacterFontVariant(text='𝗘', kind=81), CharacterFontVariant(text='𝘌', kind=88), CharacterFontVariant(text='𝙀', kind=89), CharacterFontVariant(text='𝙴', kind=48)], 'F': [CharacterFontVariant(text='ℱ', kind=128), CharacterFontVariant(text='𝐅', kind=17), CharacterFontVariant(text='𝐹', kind=24), CharacterFontVariant(text='𝑭', kind=25), CharacterFontVariant(text='𝓕', kind=145), CharacterFontVariant(text='𝔉', kind=20), CharacterFontVariant(text='𝔽', kind=18), CharacterFontVariant(text='𝕱', kind=21), CharacterFontVariant(text='𝖥', kind=80), CharacterFontVariant(text='𝗙', kind=81), CharacterFontVariant(text='𝘍', kind=88), CharacterFontVariant(text='𝙁', kind=89), CharacterFontVariant(text='𝙵', kind=48)], 'M': [CharacterFontVariant(text='ℳ', kind=128), CharacterFontVariant(text='𝐌', kind=17), CharacterFontVariant(text='𝑀', kind=24), CharacterFontVariant(text='𝑴', kind=25), CharacterFontVariant(text='𝓜', kind=145), CharacterFontVariant(text='𝔐', kind=20), CharacterFontVariant(text='𝕄', kind=18), CharacterFontVariant(text='𝕸', kind=21), CharacterFontVariant(text='𝖬', kind=80), CharacterFontVariant(text='𝗠', kind=81), CharacterFontVariant(text='𝘔', kind=88), CharacterFontVariant(text='𝙈', kind=89), CharacterFontVariant(text='𝙼', kind=48)], 'o': [CharacterFontVariant(text='ℴ', kind=128), CharacterFontVariant(text='𝐨', kind=17), CharacterFontVariant(text='𝑜', kind=24), CharacterFontVariant(text='𝒐', kind=25), CharacterFontVariant(text='𝓸', kind=145), CharacterFontVariant(text='𝔬', kind=20), CharacterFontVariant(text='𝕠', kind=18), CharacterFontVariant(text='𝖔', kind=21), CharacterFontVariant(text='𝗈', kind=80), CharacterFontVariant(text='𝗼', kind=81), CharacterFontVariant(text='𝘰', kind=88), CharacterFontVariant(text='𝙤', kind=89), CharacterFontVariant(text='𝚘', kind=48)], 'i': [CharacterFontVariant(text='ℹ', kind=0), CharacterFontVariant(text='ⅈ', kind=10), CharacterFontVariant(text='𝐢', kind=17), CharacterFontVariant(text='𝑖', kind=24), CharacterFontVariant(text='𝒊', kind=25), CharacterFontVariant(text='𝒾', kind=144), CharacterFontVariant(text='𝓲', kind=145), CharacterFontVariant(text='𝔦', kind=20), CharacterFontVariant(text='𝕚', kind=18), CharacterFontVariant(text='𝖎', kind=21), CharacterFontVariant(text='𝗂', kind=80), CharacterFontVariant(text='𝗶', kind=81), CharacterFontVariant(text='𝘪', kind=88), CharacterFontVariant(text='𝙞', kind=89), CharacterFontVariant(text='𝚒', kind=48)], 'π': [CharacterFontVariant(text='ℼ', kind=2), CharacterFontVariant(text='𝛑', kind=17), CharacterFontVariant(text='𝜋', kind=24), CharacterFontVariant(text='𝝅', kind=25), CharacterFontVariant(text='𝝿', kind=81), CharacterFontVariant(text='𝞹', kind=89)], 'γ': [CharacterFontVariant(text='ℽ', kind=2), CharacterFontVariant(text='𝛄', kind=17), CharacterFontVariant(text='𝛾', kind=24), CharacterFontVariant(text='𝜸', kind=25), CharacterFontVariant(text='𝝲', kind=81), CharacterFontVariant(text='𝞬', kind=89)], 'Γ': [CharacterFontVariant(text='ℾ', kind=2), CharacterFontVariant(text='𝚪', kind=17), CharacterFontVariant(text='𝛤', kind=24), CharacterFontVariant(text='𝜞', kind=25), CharacterFontVariant(text='𝝘', kind=81), CharacterFontVariant(text='𝞒', kind=89)], 'Π': [CharacterFontVariant(text='ℿ', kind=2), CharacterFontVariant(text='𝚷', kind=17), CharacterFontVariant(text='𝛱', kind=24), CharacterFontVariant(text='𝜫', kind=25), CharacterFontVariant(text='𝝥', kind=81), CharacterFontVariant(text='𝞟', kind=89)], '∑': [CharacterFontVariant(text='⅀', kind=2)], 'D': [CharacterFontVariant(text='ⅅ', kind=10), CharacterFontVariant(text='𝐃', kind=17), CharacterFontVariant(text='𝐷', kind=24), CharacterFontVariant(text='𝑫', kind=25), CharacterFontVariant(text='𝒟', kind=144), CharacterFontVariant(text='𝓓', kind=145), CharacterFontVariant(text='𝔇', kind=20), CharacterFontVariant(text='𝔻', kind=18), CharacterFontVariant(text='𝕯', kind=21), CharacterFontVariant(text='𝖣', kind=80), CharacterFontVariant(text='𝗗', kind=81), CharacterFontVariant(text='𝘋', kind=88), CharacterFontVariant(text='𝘿', kind=89), CharacterFontVariant(text='𝙳', kind=48)], 'd': [CharacterFontVariant(text='ⅆ', kind=10), CharacterFontVariant(text='𝐝', kind=17), CharacterFontVariant(text='𝑑', kind=24), CharacterFontVariant(text='𝒅', kind=25), CharacterFontVariant(text='𝒹', kind=144), CharacterFontVariant(text='𝓭', kind=145), CharacterFontVariant(text='𝔡', kind=20), CharacterFontVariant(text='𝕕', kind=18), CharacterFontVariant(text='𝖉', kind=21), CharacterFontVariant(text='𝖽', kind=80), CharacterFontVariant(text='𝗱', kind=81), CharacterFontVariant(text='𝘥', kind=88), CharacterFontVariant(text='𝙙', kind=89), CharacterFontVariant(text='𝚍', kind=48)], 'j': [CharacterFontVariant(text='ⅉ', kind=10), CharacterFontVariant(text='𝐣', kind=17), CharacterFontVariant(text='𝑗', kind=24), CharacterFontVariant(text='𝒋', kind=25), CharacterFontVariant(text='𝒿', kind=144), CharacterFontVariant(text='𝓳', kind=145), CharacterFontVariant(text='𝔧', kind=20), CharacterFontVariant(text='𝕛', kind=18), CharacterFontVariant(text='𝖏', kind=21), CharacterFontVariant(text='𝗃', kind=80), CharacterFontVariant(text='𝗷', kind=81), CharacterFontVariant(text='𝘫', kind=88), CharacterFontVariant(text='𝙟', kind=89), CharacterFontVariant(text='𝚓', kind=48)], 'ע': [CharacterFontVariant(text='ﬠ', kind=0)], 'א': [CharacterFontVariant(text='ﬡ', kind=0)], 'ד': [CharacterFontVariant(text='ﬢ', kind=0)], 'ה': [CharacterFontVariant(text='ﬣ', kind=0)], 'כ': [CharacterFontVariant(text='ﬤ', kind=0)], 'ל': [CharacterFontVariant(text='ﬥ', kind=0)], 'ם': [CharacterFontVariant(text='ﬦ', kind=0)], 'ר': [CharacterFontVariant(text='ﬧ', kind=0)], 'ת': [CharacterFontVariant(text='ﬨ', kind=0)], '+': [CharacterFontVariant(text='﬩', kind=0)], 'A': [CharacterFontVariant(text='𝐀', kind=17), CharacterFontVariant(text='𝐴', kind=24), CharacterFontVariant(text='𝑨', kind=25), CharacterFontVariant(text='𝒜', kind=144), CharacterFontVariant(text='𝓐', kind=145), CharacterFontVariant(text='𝔄', kind=20), CharacterFontVariant(text='𝔸', kind=18), CharacterFontVariant(text='𝕬', kind=21), CharacterFontVariant(text='𝖠', kind=80), CharacterFontVariant(text='𝗔', kind=81), CharacterFontVariant(text='𝘈', kind=88), CharacterFontVariant(text='𝘼', kind=89), CharacterFontVariant(text='𝙰', kind=48)], 'G': [CharacterFontVariant(text='𝐆', kind=17), CharacterFontVariant(text='𝐺', kind=24), CharacterFontVariant(text='𝑮', kind=25), CharacterFontVariant(text='𝒢', kind=144), CharacterFontVariant(text='𝓖', kind=145), CharacterFontVariant(text='𝔊', kind=20), CharacterFontVariant(text='𝔾', kind=18), CharacterFontVariant(text='𝕲', kind=21), CharacterFontVariant(text='𝖦', kind=80), CharacterFontVariant(text='𝗚', kind=81), CharacterFontVariant(text='𝘎', kind=88), CharacterFontVariant(text='𝙂', kind=89), CharacterFontVariant(text='𝙶', kind=48)], 'J': [CharacterFontVariant(text='𝐉', kind=17), CharacterFontVariant(text='𝐽', kind=24), CharacterFontVariant(text='𝑱', kind=25), CharacterFontVariant(text='𝒥', kind=144), CharacterFontVariant(text='𝓙', kind=145), CharacterFontVariant(text='𝔍', kind=20), CharacterFontVariant(text='𝕁', kind=18), CharacterFontVariant(text='𝕵', kind=21), CharacterFontVariant(text='𝖩', kind=80), CharacterFontVariant(text='𝗝', kind=81), CharacterFontVariant(text='𝘑', kind=88), CharacterFontVariant(text='𝙅', kind=89), CharacterFontVariant(text='𝙹', kind=48)], 'K': [CharacterFontVariant(text='𝐊', kind=17), CharacterFontVariant(text='𝐾', kind=24), CharacterFontVariant(text='𝑲', kind=25), CharacterFontVariant(text='𝒦', kind=144), CharacterFontVariant(text='𝓚', kind=145), CharacterFontVariant(text='𝔎', kind=20), CharacterFontVariant(text='𝕂', kind=18), CharacterFontVariant(text='𝕶', kind=21), CharacterFontVariant(text='𝖪', kind=80), CharacterFontVariant(text='𝗞', kind=81), CharacterFontVariant(text='𝘒', kind=88), CharacterFontVariant(text='𝙆', kind=89), CharacterFontVariant(text='𝙺', kind=48)], 'O': [CharacterFontVariant(text='𝐎', kind=17), CharacterFontVariant(text='𝑂', kind=24), CharacterFontVariant(text='𝑶', kind=25), CharacterFontVariant(text='𝒪', kind=144), CharacterFontVariant(text='𝓞', kind=145), CharacterFontVariant(text='𝔒', kind=20), CharacterFontVariant(text='𝕆', kind=18), CharacterFontVariant(text='𝕺', kind=21), CharacterFontVariant(text='𝖮', kind=80), CharacterFontVariant(text='𝗢', kind=81), CharacterFontVariant(text='𝘖', kind=88), CharacterFontVariant(text='𝙊', kind=89), CharacterFontVariant(text='𝙾', kind=48)], 'S': [CharacterFontVariant(text='𝐒', kind=17), CharacterFontVariant(text='𝑆', kind=24), CharacterFontVariant(text='𝑺', kind=25), CharacterFontVariant(text='𝒮', kind=144), CharacterFontVariant(text='𝓢', kind=145), CharacterFontVariant(text='𝔖', kind=20), CharacterFontVariant(text='𝕊', kind=18), CharacterFontVariant(text='𝕾', kind=21), CharacterFontVariant(text='𝖲', kind=80), CharacterFontVariant(text='𝗦', kind=81), CharacterFontVariant(text='𝘚', kind=88), CharacterFontVariant(text='𝙎', kind=89), CharacterFontVariant(text='𝚂', kind=48)], 'T': [CharacterFontVariant(text='𝐓', kind=17), CharacterFontVariant(text='𝑇', kind=24), CharacterFontVariant(text='𝑻', kind=25), CharacterFontVariant(text='𝒯', kind=144), CharacterFontVariant(text='𝓣', kind=145), CharacterFontVariant(text='𝔗', kind=20), CharacterFontVariant(text='𝕋', kind=18), CharacterFontVariant(text='𝕿', kind=21), CharacterFontVariant(text='𝖳', kind=80), CharacterFontVariant(text='𝗧', kind=81), CharacterFontVariant(text='𝘛', kind=88), CharacterFontVariant(text='𝙏', kind=89), CharacterFontVariant(text='𝚃', kind=48)], 'U': [CharacterFontVariant(text='𝐔', kind=17), CharacterFontVariant(text='𝑈', kind=24), CharacterFontVariant(text='𝑼', kind=25), CharacterFontVariant(text='𝒰', kind=144), CharacterFontVariant(text='𝓤', kind=145), CharacterFontVariant(text='𝔘', kind=20), CharacterFontVariant(text='𝕌', kind=18), CharacterFontVariant(text='𝖀', kind=21), CharacterFontVariant(text='𝖴', kind=80), CharacterFontVariant(text='𝗨', kind=81), CharacterFontVariant(text='𝘜', kind=88), CharacterFontVariant(text='𝙐', kind=89), CharacterFontVariant(text='𝚄', kind=48)], 'V': [CharacterFontVariant(text='𝐕', kind=17), CharacterFontVariant(text='𝑉', kind=24), CharacterFontVariant(text='𝑽', kind=25), CharacterFontVariant(text='𝒱', kind=144), CharacterFontVariant(text='𝓥', kind=145), CharacterFontVariant(text='𝔙', kind=20), CharacterFontVariant(text='𝕍', kind=18), CharacterFontVariant(text='𝖁', kind=21), CharacterFontVariant(text='𝖵', kind=80), CharacterFontVariant(text='𝗩', kind=81), CharacterFontVariant(text='𝘝', kind=88), CharacterFontVariant(text='𝙑', kind=89), CharacterFontVariant(text='𝚅', kind=48)], 'W': [CharacterFontVariant(text='𝐖', kind=17), CharacterFontVariant(text='𝑊', kind=24), CharacterFontVariant(text='𝑾', kind=25), CharacterFontVariant(text='𝒲', kind=144), CharacterFontVariant(text='𝓦', kind=145), CharacterFontVariant(text='𝔚', kind=20), CharacterFontVariant(text='𝕎', kind=18), CharacterFontVariant(text='𝖂', kind=21), CharacterFontVariant(text='𝖶', kind=80), CharacterFontVariant(text='𝗪', kind=81), CharacterFontVariant(text='𝘞', kind=88), CharacterFontVariant(text='𝙒', kind=89), CharacterFontVariant(text='𝚆', kind=48)], 'X': [CharacterFontVariant(text='𝐗', kind=17), CharacterFontVariant(text='𝑋', kind=24), CharacterFontVariant(text='𝑿', kind=25), CharacterFontVariant(text='𝒳', kind=144), CharacterFontVariant(text='𝓧', kind=145), CharacterFontVariant(text='𝔛', kind=20), CharacterFontVariant(text='𝕏', kind=18), CharacterFontVariant(text='𝖃', kind=21), CharacterFontVariant(text='𝖷', kind=80), CharacterFontVariant(text='𝗫', kind=81), CharacterFontVariant(text='𝘟', kind=88), CharacterFontVariant(text='𝙓', kind=89), CharacterFontVariant(text='𝚇', kind=48)], 'Y': [CharacterFontVariant(text='𝐘', kind=17), CharacterFontVariant(text='𝑌', kind=24), CharacterFontVariant(text='𝒀', kind=25), CharacterFontVariant(text='𝒴', kind=144), CharacterFontVariant(text='𝓨', kind=145), CharacterFontVariant(text='𝔜', kind=20), CharacterFontVariant(text='𝕐', kind=18), CharacterFontVariant(text='𝖄', kind=21), CharacterFontVariant(text='𝖸', kind=80), CharacterFontVariant(text='𝗬', kind=81), CharacterFontVariant(text='𝘠', kind=88), CharacterFontVariant(text='𝙔', kind=89), CharacterFontVariant(text='𝚈', kind=48)], 'a': [CharacterFontVariant(text='𝐚', kind=17), CharacterFontVariant(text='𝑎', kind=24), CharacterFontVariant(text='𝒂', kind=25), CharacterFontVariant(text='𝒶', kind=144), CharacterFontVariant(text='𝓪', kind=145), CharacterFontVariant(text='𝔞', kind=20), CharacterFontVariant(text='𝕒', kind=18), CharacterFontVariant(text='𝖆', kind=21), CharacterFontVariant(text='𝖺', kind=80), CharacterFontVariant(text='𝗮', kind=81), CharacterFontVariant(text='𝘢', kind=88), CharacterFontVariant(text='𝙖', kind=89), CharacterFontVariant(text='𝚊', kind=48)], 'b': [CharacterFontVariant(text='𝐛', kind=17), CharacterFontVariant(text='𝑏', kind=24), CharacterFontVariant(text='𝒃', kind=25), CharacterFontVariant(text='𝒷', kind=144), CharacterFontVariant(text='𝓫', kind=145), CharacterFontVariant(text='𝔟', kind=20), CharacterFontVariant(text='𝕓', kind=18), CharacterFontVariant(text='𝖇', kind=21), CharacterFontVariant(text='𝖻', kind=80), CharacterFontVariant(text='𝗯', kind=81), CharacterFontVariant(text='𝘣', kind=88), CharacterFontVariant(text='𝙗', kind=89), CharacterFontVariant(text='𝚋', kind=48)], 'c': [CharacterFontVariant(text='𝐜', kind=17), CharacterFontVariant(text='𝑐', kind=24), CharacterFontVariant(text='𝒄', kind=25), CharacterFontVariant(text='𝒸', kind=144), CharacterFontVariant(text='𝓬', kind=145), CharacterFontVariant(text='𝔠', kind=20), CharacterFontVariant(text='𝕔', kind=18), CharacterFontVariant(text='𝖈', kind=21), CharacterFontVariant(text='𝖼', kind=80), CharacterFontVariant(text='𝗰', kind=81), CharacterFontVariant(text='𝘤', kind=88), CharacterFontVariant(text='𝙘', kind=89), CharacterFontVariant(text='𝚌', kind=48)], 'f': [CharacterFontVariant(text='𝐟', kind=17), CharacterFontVariant(text='𝑓', kind=24), CharacterFontVariant(text='𝒇', kind=25), CharacterFontVariant(text='𝒻', kind=144), CharacterFontVariant(text='𝓯', kind=145), CharacterFontVariant(text='𝔣', kind=20), CharacterFontVariant(text='𝕗', kind=18), CharacterFontVariant(text='𝖋', kind=21), CharacterFontVariant(text='𝖿', kind=80), CharacterFontVariant(text='𝗳', kind=81), CharacterFontVariant(text='𝘧', kind=88), CharacterFontVariant(text='𝙛', kind=89), CharacterFontVariant(text='𝚏', kind=48)], 'k': [CharacterFontVariant(text='𝐤', kind=17), CharacterFontVariant(text='𝑘', kind=24), CharacterFontVariant(text='𝒌', kind=25), CharacterFontVariant(text='𝓀', kind=144), CharacterFontVariant(text='𝓴', kind=145), CharacterFontVariant(text='𝔨', kind=20), CharacterFontVariant(text='𝕜', kind=18), CharacterFontVariant(text='𝖐', kind=21), CharacterFontVariant(text='𝗄', kind=80), CharacterFontVariant(text='𝗸', kind=81), CharacterFontVariant(text='𝘬', kind=88), CharacterFontVariant(text='𝙠', kind=89), CharacterFontVariant(text='𝚔', kind=48)], 'm': [CharacterFontVariant(text='𝐦', kind=17), CharacterFontVariant(text='𝑚', kind=24), CharacterFontVariant(text='𝒎', kind=25), CharacterFontVariant(text='𝓂', kind=144), CharacterFontVariant(text='𝓶', kind=145), CharacterFontVariant(text='𝔪', kind=20), CharacterFontVariant(text='𝕞', kind=18), CharacterFontVariant(text='𝖒', kind=21), CharacterFontVariant(text='𝗆', kind=80), CharacterFontVariant(text='𝗺', kind=81), CharacterFontVariant(text='𝘮', kind=88), CharacterFontVariant(text='𝙢', kind=89), CharacterFontVariant(text='𝚖', kind=48)], 'n': [CharacterFontVariant(text='𝐧', kind=17), CharacterFontVariant(text='𝑛', kind=24), CharacterFontVariant(text='𝒏', kind=25), CharacterFontVariant(text='𝓃', kind=144), CharacterFontVariant(text='𝓷', kind=145), CharacterFontVariant(text='𝔫', kind=20), CharacterFontVariant(text='𝕟', kind=18), CharacterFontVariant(text='𝖓', kind=21), CharacterFontVariant(text='𝗇', kind=80), CharacterFontVariant(text='𝗻', kind=81), CharacterFontVariant(text='𝘯', kind=88), CharacterFontVariant(text='𝙣', kind=89), CharacterFontVariant(text='𝚗', kind=48)], 'p': [CharacterFontVariant(text='𝐩', kind=17), CharacterFontVariant(text='𝑝', kind=24), CharacterFontVariant(text='𝒑', kind=25), CharacterFontVariant(text='𝓅', kind=144), CharacterFontVariant(text='𝓹', kind=145), CharacterFontVariant(text='𝔭', kind=20), CharacterFontVariant(text='𝕡', kind=18), CharacterFontVariant(text='𝖕', kind=21), CharacterFontVariant(text='𝗉', kind=80), CharacterFontVariant(text='𝗽', kind=81), CharacterFontVariant(text='𝘱', kind=88), CharacterFontVariant(text='𝙥', kind=89), CharacterFontVariant(text='𝚙', kind=48)], 'q': [CharacterFontVariant(text='𝐪', kind=17), CharacterFontVariant(text='𝑞', kind=24), CharacterFontVariant(text='𝒒', kind=25), CharacterFontVariant(text='𝓆', kind=144), CharacterFontVariant(text='𝓺', kind=145), CharacterFontVariant(text='𝔮', kind=20), CharacterFontVariant(text='𝕢', kind=18), CharacterFontVariant(text='𝖖', kind=21), CharacterFontVariant(text='𝗊', kind=80), CharacterFontVariant(text='𝗾', kind=81), CharacterFontVariant(text='𝘲', kind=88), CharacterFontVariant(text='𝙦', kind=89), CharacterFontVariant(text='𝚚', kind=48)], 'r': [CharacterFontVariant(text='𝐫', kind=17), CharacterFontVariant(text='𝑟', kind=24), CharacterFontVariant(text='𝒓', kind=25), CharacterFontVariant(text='𝓇', kind=144), CharacterFontVariant(text='𝓻', kind=145), CharacterFontVariant(text='𝔯', kind=20), CharacterFontVariant(text='𝕣', kind=18), CharacterFontVariant(text='𝖗', kind=21), CharacterFontVariant(text='𝗋', kind=80), CharacterFontVariant(text='𝗿', kind=81), CharacterFontVariant(text='𝘳', kind=88), CharacterFontVariant(text='𝙧', kind=89), CharacterFontVariant(text='𝚛', kind=48)], 's': [CharacterFontVariant(text='𝐬', kind=17), CharacterFontVariant(text='𝑠', kind=24), CharacterFontVariant(text='𝒔', kind=25), CharacterFontVariant(text='𝓈', kind=144), CharacterFontVariant(text='𝓼', kind=145), CharacterFontVariant(text='𝔰', kind=20), CharacterFontVariant(text='𝕤', kind=18), CharacterFontVariant(text='𝖘', kind=21), CharacterFontVariant(text='𝗌', kind=80), CharacterFontVariant(text='𝘀', kind=81), CharacterFontVariant(text='𝘴', kind=88), CharacterFontVariant(text='𝙨', kind=89), CharacterFontVariant(text='𝚜', kind=48)], 't': [CharacterFontVariant(text='𝐭', kind=17), CharacterFontVariant(text='𝑡', kind=24), CharacterFontVariant(text='𝒕', kind=25), CharacterFontVariant(text='𝓉', kind=144), CharacterFontVariant(text='𝓽', kind=145), CharacterFontVariant(text='𝔱', kind=20), CharacterFontVariant(text='𝕥', kind=18), CharacterFontVariant(text='𝖙', kind=21), CharacterFontVariant(text='𝗍', kind=80), CharacterFontVariant(text='𝘁', kind=81), CharacterFontVariant(text='𝘵', kind=88), CharacterFontVariant(text='𝙩', kind=89), CharacterFontVariant(text='𝚝', kind=48)], 'u': [CharacterFontVariant(text='𝐮', kind=17), CharacterFontVariant(text='𝑢', kind=24), CharacterFontVariant(text='𝒖', kind=25), CharacterFontVariant(text='𝓊', kind=144), CharacterFontVariant(text='𝓾', kind=145), CharacterFontVariant(text='𝔲', kind=20), CharacterFontVariant(text='𝕦', kind=18), CharacterFontVariant(text='𝖚', kind=21), CharacterFontVariant(text='𝗎', kind=80), CharacterFontVariant(text='𝘂', kind=81), CharacterFontVariant(text='𝘶', kind=88), CharacterFontVariant(text='𝙪', kind=89), CharacterFontVariant(text='𝚞', kind=48)], 'v': [CharacterFontVariant(text='𝐯', kind=17), CharacterFontVariant(text='𝑣', kind=24), CharacterFontVariant(text='𝒗', kind=25), CharacterFontVariant(text='𝓋', kind=144), CharacterFontVariant(text='𝓿', kind=145), CharacterFontVariant(text='𝔳', kind=20), CharacterFontVariant(text='𝕧', kind=18), CharacterFontVariant(text='𝖛', kind=21), CharacterFontVariant(text='𝗏', kind=80), CharacterFontVariant(text='𝘃', kind=81), CharacterFontVariant(text='𝘷', kind=88), CharacterFontVariant(text='𝙫', kind=89), CharacterFontVariant(text='𝚟', kind=48)], 'w': [CharacterFontVariant(text='𝐰', kind=17), CharacterFontVariant(text='𝑤', kind=24), CharacterFontVariant(text='𝒘', kind=25), CharacterFontVariant(text='𝓌', kind=144), CharacterFontVariant(text='𝔀', kind=145), CharacterFontVariant(text='𝔴', kind=20), CharacterFontVariant(text='𝕨', kind=18), CharacterFontVariant(text='𝖜', kind=21), CharacterFontVariant(text='𝗐', kind=80), CharacterFontVariant(text='𝘄', kind=81), CharacterFontVariant(text='𝘸', kind=88), CharacterFontVariant(text='𝙬', kind=89), CharacterFontVariant(text='𝚠', kind=48)], 'x': [CharacterFontVariant(text='𝐱', kind=17), CharacterFontVariant(text='𝑥', kind=24), CharacterFontVariant(text='𝒙', kind=25), CharacterFontVariant(text='𝓍', kind=144), CharacterFontVariant(text='𝔁', kind=145), CharacterFontVariant(text='𝔵', kind=20), CharacterFontVariant(text='𝕩', kind=18), CharacterFontVariant(text='𝖝', kind=21), CharacterFontVariant(text='𝗑', kind=80), CharacterFontVariant(text='𝘅', kind=81), CharacterFontVariant(text='𝘹', kind=88), CharacterFontVariant(text='𝙭', kind=89), CharacterFontVariant(text='𝚡', kind=48)], 'y': [CharacterFontVariant(text='𝐲', kind=17), CharacterFontVariant(text='𝑦', kind=24), CharacterFontVariant(text='𝒚', kind=25), CharacterFontVariant(text='𝓎', kind=144), CharacterFontVariant(text='𝔂', kind=145), CharacterFontVariant(text='𝔶', kind=20), CharacterFontVariant(text='𝕪', kind=18), CharacterFontVariant(text='𝖞', kind=21), CharacterFontVariant(text='𝗒', kind=80), CharacterFontVariant(text='𝘆', kind=81), CharacterFontVariant(text='𝘺', kind=88), CharacterFontVariant(text='𝙮', kind=89), CharacterFontVariant(text='𝚢', kind=48)], 'z': [CharacterFontVariant(text='𝐳', kind=17), CharacterFontVariant(text='𝑧', kind=24), CharacterFontVariant(text='𝒛', kind=25), CharacterFontVariant(text='𝓏', kind=144), CharacterFontVariant(text='𝔃', kind=145), CharacterFontVariant(text='𝔷', kind=20), CharacterFontVariant(text='𝕫', kind=18), CharacterFontVariant(text='𝖟', kind=21), CharacterFontVariant(text='𝗓', kind=80), CharacterFontVariant(text='𝘇', kind=81), CharacterFontVariant(text='𝘻', kind=88), CharacterFontVariant(text='𝙯', kind=89), CharacterFontVariant(text='𝚣', kind=48)], 'ı': [CharacterFontVariant(text='𝚤', kind=24)], 'ȷ': [CharacterFontVariant(text='𝚥', kind=24)], 'Α': [CharacterFontVariant(text='𝚨', kind=17), CharacterFontVariant(text='𝛢', kind=24), CharacterFontVariant(text='𝜜', kind=25), CharacterFontVariant(text='𝝖', kind=81), CharacterFontVariant(text='𝞐', kind=89)], 'Β': [CharacterFontVariant(text='𝚩', kind=17), CharacterFontVariant(text='𝛣', kind=24), CharacterFontVariant(text='𝜝', kind=25), CharacterFontVariant(text='𝝗', kind=81), CharacterFontVariant(text='𝞑', kind=89)], 'Δ': [CharacterFontVariant(text='𝚫', kind=17), CharacterFontVariant(text='𝛥', kind=24), CharacterFontVariant(text='𝜟', kind=25), CharacterFontVariant(text='𝝙', kind=81), CharacterFontVariant(text='𝞓', kind=89)], 'Ε': [CharacterFontVariant(text='𝚬', kind=17), CharacterFontVariant(text='𝛦', kind=24), CharacterFontVariant(text='𝜠', kind=25), CharacterFontVariant(text='𝝚', kind=81), CharacterFontVariant(text='𝞔', kind=89)], 'Ζ': [CharacterFontVariant(text='𝚭', kind=17), CharacterFontVariant(text='𝛧', kind=24), CharacterFontVariant(text='𝜡', kind=25), CharacterFontVariant(text='𝝛', kind=81), CharacterFontVariant(text='𝞕', kind=89)], 'Η': [CharacterFontVariant(text='𝚮', kind=17), CharacterFontVariant(text='𝛨', kind=24), CharacterFontVariant(text='𝜢', kind=25), CharacterFontVariant(text='𝝜', kind=81), CharacterFontVariant(text='𝞖', kind=89)], 'Θ': [CharacterFontVariant(text='𝚯', kind=17), CharacterFontVariant(text='𝛩', kind=24), CharacterFontVariant(text='𝜣', kind=25), CharacterFontVariant(text='𝝝', kind=81), CharacterFontVariant(text='𝞗', kind=89)], 'Ι': [CharacterFontVariant(text='𝚰', kind=17), CharacterFontVariant(text='𝛪', kind=24), CharacterFontVariant(text='𝜤', kind=25), CharacterFontVariant(text='𝝞', kind=81), CharacterFontVariant(text='𝞘', kind=89)], 'Κ': [CharacterFontVariant(text='𝚱', kind=17), CharacterFontVariant(text='𝛫', kind=24), CharacterFontVariant(text='𝜥', kind=25), CharacterFontVariant(text='𝝟', kind=81), CharacterFontVariant(text='𝞙', kind=89)], 'Λ': [CharacterFontVariant(text='𝚲', kind=17), CharacterFontVariant(text='𝛬', kind=24), CharacterFontVariant(text='𝜦', kind=25), CharacterFontVariant(text='𝝠', kind=81), CharacterFontVariant(text='𝞚', kind=89)], 'Μ': [CharacterFontVariant(text='𝚳', kind=17), CharacterFontVariant(text='𝛭', kind=24), CharacterFontVariant(text='𝜧', kind=25), CharacterFontVariant(text='𝝡', kind=81), CharacterFontVariant(text='𝞛', kind=89)], 'Ν': [CharacterFontVariant(text='𝚴', kind=17), CharacterFontVariant(text='𝛮', kind=24), CharacterFontVariant(text='𝜨', kind=25), CharacterFontVariant(text='𝝢', kind=81), CharacterFontVariant(text='𝞜', kind=89)], 'Ξ': [CharacterFontVariant(text='𝚵', kind=17), CharacterFontVariant(text='𝛯', kind=24), CharacterFontVariant(text='𝜩', kind=25), CharacterFontVariant(text='𝝣', kind=81), CharacterFontVariant(text='𝞝', kind=89)], 'Ο': [CharacterFontVariant(text='𝚶', kind=17), CharacterFontVariant(text='𝛰', kind=24), CharacterFontVariant(text='𝜪', kind=25), CharacterFontVariant(text='𝝤', kind=81), CharacterFontVariant(text='𝞞', kind=89)], 'Ρ': [CharacterFontVariant(text='𝚸', kind=17), CharacterFontVariant(text='𝛲', kind=24), CharacterFontVariant(text='𝜬', kind=25), CharacterFontVariant(text='𝝦', kind=81), CharacterFontVariant(text='𝞠', kind=89)], 'ϴ': [CharacterFontVariant(text='𝚹', kind=17), CharacterFontVariant(text='𝛳', kind=24), CharacterFontVariant(text='𝜭', kind=25), CharacterFontVariant(text='𝝧', kind=81), CharacterFontVariant(text='𝞡', kind=89)], 'Σ': [CharacterFontVariant(text='𝚺', kind=17), CharacterFontVariant(text='𝛴', kind=24), CharacterFontVariant(text='𝜮', kind=25), CharacterFontVariant(text='𝝨', kind=81), CharacterFontVariant(text='𝞢', kind=89)], 'Τ': [CharacterFontVariant(text='𝚻', kind=17), CharacterFontVariant(text='𝛵', kind=24), CharacterFontVariant(text='𝜯', kind=25), CharacterFontVariant(text='𝝩', kind=81), CharacterFontVariant(text='𝞣', kind=89)], 'Υ': [CharacterFontVariant(text='𝚼', kind=17), CharacterFontVariant(text='𝛶', kind=24), CharacterFontVariant(text='𝜰', kind=25), CharacterFontVariant(text='𝝪', kind=81), CharacterFontVariant(text='𝞤', kind=89)], 'Φ': [CharacterFontVariant(text='𝚽', kind=17), CharacterFontVariant(text='𝛷', kind=24), CharacterFontVariant(text='𝜱', kind=25), CharacterFontVariant(text='𝝫', kind=81), CharacterFontVariant(text='𝞥', kind=89)], 'Χ': [CharacterFontVariant(text='𝚾', kind=17), CharacterFontVariant(text='𝛸', kind=24), CharacterFontVariant(text='𝜲', kind=25), CharacterFontVariant(text='𝝬', kind=81), CharacterFontVariant(text='𝞦', kind=89)], 'Ψ': [CharacterFontVariant(text='𝚿', kind=17), CharacterFontVariant(text='𝛹', kind=24), CharacterFontVariant(text='𝜳', kind=25), CharacterFontVariant(text='𝝭', kind=81), CharacterFontVariant(text='𝞧', kind=89)], 'Ω': [CharacterFontVariant(text='𝛀', kind=17), CharacterFontVariant(text='𝛺', kind=24), CharacterFontVariant(text='𝜴', kind=25), CharacterFontVariant(text='𝝮', kind=81), CharacterFontVariant(text='𝞨', kind=89)], '∇': [CharacterFontVariant(text='𝛁', kind=17), CharacterFontVariant(text='𝛻', kind=24), CharacterFontVariant(text='𝜵', kind=25), CharacterFontVariant(text='𝝯', kind=81), CharacterFontVariant(text='𝞩', kind=89)], 'α': [CharacterFontVariant(text='𝛂', kind=17), CharacterFontVariant(text='𝛼', kind=24), CharacterFontVariant(text='𝜶', kind=25), CharacterFontVariant(text='𝝰', kind=81), CharacterFontVariant(text='𝞪', kind=89)], 'β': [CharacterFontVariant(text='𝛃', kind=17), CharacterFontVariant(text='𝛽', kind=24), CharacterFontVariant(text='𝜷', kind=25), CharacterFontVariant(text='𝝱', kind=81), CharacterFontVariant(text='𝞫', kind=89)], 'δ': [CharacterFontVariant(text='𝛅', kind=17), CharacterFontVariant(text='𝛿', kind=24), CharacterFontVariant(text='𝜹', kind=25), CharacterFontVariant(text='𝝳', kind=81), CharacterFontVariant(text='𝞭', kind=89)], 'ε': [CharacterFontVariant(text='𝛆', kind=17), CharacterFontVariant(text='𝜀', kind=24), CharacterFontVariant(text='𝜺', kind=25), CharacterFontVariant(text='𝝴', kind=81), CharacterFontVariant(text='𝞮', kind=89)], 'ζ': [CharacterFontVariant(text='𝛇', kind=17), CharacterFontVariant(text='𝜁', kind=24), CharacterFontVariant(text='𝜻', kind=25), CharacterFontVariant(text='𝝵', kind=81), CharacterFontVariant(text='𝞯', kind=89)], 'η': [CharacterFontVariant(text='𝛈', kind=17), CharacterFontVariant(text='𝜂', kind=24), CharacterFontVariant(text='𝜼', kind=25), CharacterFontVariant(text='𝝶', kind=81), CharacterFontVariant(text='𝞰', kind=89)], 'θ': [CharacterFontVariant(text='𝛉', kind=17), CharacterFontVariant(text='𝜃', kind=24), CharacterFontVariant(text='𝜽', kind=25), CharacterFontVariant(text='𝝷', kind=81), CharacterFontVariant(text='𝞱', kind=89)], 'ι': [CharacterFontVariant(text='𝛊', kind=17), CharacterFontVariant(text='𝜄', kind=24), CharacterFontVariant(text='𝜾', kind=25), CharacterFontVariant(text='𝝸', kind=81), CharacterFontVariant(text='𝞲', kind=89)], 'κ': [CharacterFontVariant(text='𝛋', kind=17), CharacterFontVariant(text='𝜅', kind=24), CharacterFontVariant(text='𝜿', kind=25), CharacterFontVariant(text='𝝹', kind=81), CharacterFontVariant(text='𝞳', kind=89)], 'λ': [CharacterFontVariant(text='𝛌', kind=17), CharacterFontVariant(text='𝜆', kind=24), CharacterFontVariant(text='𝝀', kind=25), CharacterFontVariant(text='𝝺', kind=81), CharacterFontVariant(text='𝞴', kind=89)], 'μ': [CharacterFontVariant(text='𝛍', kind=17), CharacterFontVariant(text='𝜇', kind=24), CharacterFontVariant(text='𝝁', kind=25), CharacterFontVariant(text='𝝻', kind=81), CharacterFontVariant(text='𝞵', kind=89)], 'ν': [CharacterFontVariant(text='𝛎', kind=17), CharacterFontVariant(text='𝜈', kind=24), CharacterFontVariant(text='𝝂', kind=25), CharacterFontVariant(text='𝝼', kind=81), CharacterFontVariant(text='𝞶', kind=89)], 'ξ': [CharacterFontVariant(text='𝛏', kind=17), CharacterFontVariant(text='𝜉', kind=24), CharacterFontVariant(text='𝝃', kind=25), CharacterFontVariant(text='𝝽', kind=81), CharacterFontVariant(text='𝞷', kind=89)], 'ο': [CharacterFontVariant(text='𝛐', kind=17), CharacterFontVariant(text='𝜊', kind=24), CharacterFontVariant(text='𝝄', kind=25), CharacterFontVariant(text='𝝾', kind=81), CharacterFontVariant(text='𝞸', kind=89)], 'ρ': [CharacterFontVariant(text='𝛒', kind=17), CharacterFontVariant(text='𝜌', kind=24), CharacterFontVariant(text='𝝆', kind=25), CharacterFontVariant(text='𝞀', kind=81), CharacterFontVariant(text='𝞺', kind=89)], 'ς': [CharacterFontVariant(text='𝛓', kind=17), CharacterFontVariant(text='𝜍', kind=24), CharacterFontVariant(text='𝝇', kind=25), CharacterFontVariant(text='𝞁', kind=81), CharacterFontVariant(text='𝞻', kind=89)], 'σ': [CharacterFontVariant(text='𝛔', kind=17), CharacterFontVariant(text='𝜎', kind=24), CharacterFontVariant(text='𝝈', kind=25), CharacterFontVariant(text='𝞂', kind=81), CharacterFontVariant(text='𝞼', kind=89)], 'τ': [CharacterFontVariant(text='𝛕', kind=17), CharacterFontVariant(text='𝜏', kind=24), CharacterFontVariant(text='𝝉', kind=25), CharacterFontVariant(text='𝞃', kind=81), CharacterFontVariant(text='𝞽', kind=89)], 'υ': [CharacterFontVariant(text='𝛖', kind=17), CharacterFontVariant(text='𝜐', kind=24), CharacterFontVariant(text='𝝊', kind=25), CharacterFontVariant(text='𝞄', kind=81), CharacterFontVariant(text='𝞾', kind=89)], 'φ': [CharacterFontVariant(text='𝛗', kind=17), CharacterFontVariant(text='𝜑', kind=24), CharacterFontVariant(text='𝝋', kind=25), CharacterFontVariant(text='𝞅', kind=81), CharacterFontVariant(text='𝞿', kind=89)], 'χ': [CharacterFontVariant(text='𝛘', kind=17), CharacterFontVariant(text='𝜒', kind=24), CharacterFontVariant(text='𝝌', kind=25), CharacterFontVariant(text='𝞆', kind=81), CharacterFontVariant(text='𝟀', kind=89)], 'ψ': [CharacterFontVariant(text='𝛙', kind=17), CharacterFontVariant(text='𝜓', kind=24), CharacterFontVariant(text='𝝍', kind=25), CharacterFontVariant(text='𝞇', kind=81), CharacterFontVariant(text='𝟁', kind=89)], 'ω': [CharacterFontVariant(text='𝛚', kind=17), CharacterFontVariant(text='𝜔', kind=24), CharacterFontVariant(text='𝝎', kind=25), CharacterFontVariant(text='𝞈', kind=81), CharacterFontVariant(text='𝟂', kind=89)], '∂': [CharacterFontVariant(text='𝛛', kind=17), CharacterFontVariant(text='𝜕', kind=24), CharacterFontVariant(text='𝝏', kind=25), CharacterFontVariant(text='𝞉', kind=81), CharacterFontVariant(text='𝟃', kind=89)], 'ϵ': [CharacterFontVariant(text='𝛜', kind=17), CharacterFontVariant(text='𝜖', kind=24), CharacterFontVariant(text='𝝐', kind=25), CharacterFontVariant(text='𝞊', kind=81), CharacterFontVariant(text='𝟄', kind=89)], 'ϑ': [CharacterFontVariant(text='𝛝', kind=17), CharacterFontVariant(text='𝜗', kind=24), CharacterFontVariant(text='𝝑', kind=25), CharacterFontVariant(text='𝞋', kind=81), CharacterFontVariant(text='𝟅', kind=89)], 'ϰ': [CharacterFontVariant(text='𝛞', kind=17), CharacterFontVariant(text='𝜘', kind=24), CharacterFontVariant(text='𝝒', kind=25), CharacterFontVariant(text='𝞌', kind=81), CharacterFontVariant(text='𝟆', kind=89)], 'ϕ': [CharacterFontVariant(text='𝛟', kind=17), CharacterFontVariant(text='𝜙', kind=24), CharacterFontVariant(text='𝝓', kind=25), CharacterFontVariant(text='𝞍', kind=81), CharacterFontVariant(text='𝟇', kind=89)], 'ϱ': [CharacterFontVariant(text='𝛠', kind=17), CharacterFontVariant(text='𝜚', kind=24), CharacterFontVariant(text='𝝔', kind=25), CharacterFontVariant(text='𝞎', kind=81), CharacterFontVariant(text='𝟈', kind=89)], 'ϖ': [CharacterFontVariant(text='𝛡', kind=17), CharacterFontVariant(text='𝜛', kind=24), CharacterFontVariant(text='𝝕', kind=25), CharacterFontVariant(text='𝞏', kind=81), CharacterFontVariant(text='𝟉', kind=89)], 'Ϝ': [CharacterFontVariant(text='𝟊', kind=17)], 'ϝ': [CharacterFontVariant(text='𝟋', kind=17)], '0': [CharacterFontVariant(text='𝟎', kind=17), CharacterFontVariant(text='𝟘', kind=18), CharacterFontVariant(text='𝟢', kind=80), CharacterFontVariant(text='𝟬', kind=81), CharacterFontVariant(text='𝟶', kind=48), CharacterFontVariant(text='🯰', kind=0)], '1': [CharacterFontVariant(text='𝟏', kind=17), CharacterFontVariant(text='𝟙', kind=18), CharacterFontVariant(text='𝟣', kind=80), CharacterFontVariant(text='𝟭', kind=81), CharacterFontVariant(text='𝟷', kind=48), CharacterFontVariant(text='🯱', kind=0)], '2': [CharacterFontVariant(text='𝟐', kind=17), CharacterFontVariant(text='𝟚', kind=18), CharacterFontVariant(text='𝟤', kind=80), CharacterFontVariant(text='𝟮', kind=81), CharacterFontVariant(text='𝟸', kind=48), CharacterFontVariant(text='🯲', kind=0)], '3': [CharacterFontVariant(text='𝟑', kind=17), CharacterFontVariant(text='𝟛', kind=18), CharacterFontVariant(text='𝟥', kind=80), CharacterFontVariant(text='𝟯', kind=81), CharacterFontVariant(text='𝟹', kind=48), CharacterFontVariant(text='🯳', kind=0)], '4': [CharacterFontVariant(text='𝟒', kind=17), CharacterFontVariant(text='𝟜', kind=18), CharacterFontVariant(text='𝟦', kind=80), CharacterFontVariant(text='𝟰', kind=81), CharacterFontVariant(text='𝟺', kind=48), CharacterFontVariant(text='🯴', kind=0)], '5': [CharacterFontVariant(text='𝟓', kind=17), CharacterFontVariant(text='𝟝', kind=18), CharacterFontVariant(text='𝟧', kind=80), CharacterFontVariant(text='𝟱', kind=81), CharacterFontVariant(text='𝟻', kind=48), CharacterFontVariant(text='🯵', kind=0)], '6': [CharacterFontVariant(text='𝟔', kind=17), CharacterFontVariant(text='𝟞', kind=18), CharacterFontVariant(text='𝟨', kind=80), CharacterFontVariant(text='𝟲', kind=81), CharacterFontVariant(text='𝟼', kind=48), CharacterFontVariant(text='🯶', kind=0)], '7': [CharacterFontVariant(text='𝟕', kind=17), CharacterFontVariant(text='𝟟', kind=18), CharacterFontVariant(text='𝟩', kind=80), CharacterFontVariant(text='𝟳', kind=81), CharacterFontVariant(text='𝟽', kind=48), CharacterFontVariant(text='🯷', kind=0)], '8': [CharacterFontVariant(text='𝟖', kind=17), CharacterFontVariant(text='𝟠', kind=18), CharacterFontVariant(text='𝟪', kind=80), CharacterFontVariant(text='𝟴', kind=81), CharacterFontVariant(text='𝟾', kind=48), CharacterFontVariant(text='🯸', kind=0)], '9': [CharacterFontVariant(text='𝟗', kind=17), CharacterFontVariant(text='𝟡', kind=18), CharacterFontVariant(text='𝟫', kind=80), CharacterFontVariant(text='𝟵', kind=81), CharacterFontVariant(text='𝟿', kind=48), CharacterFontVariant(text='🯹', kind=0)], 'ا': [CharacterFontVariant(text='𞸀', kind=16), CharacterFontVariant(text='𞺀', kind=16)], 'ب': [CharacterFontVariant(text='𞸁', kind=16), CharacterFontVariant(text='𞸡', kind=16), CharacterFontVariant(text='𞹡', kind=16), CharacterFontVariant(text='𞺁', kind=16), CharacterFontVariant(text='𞺡', kind=18)], 'ج': [CharacterFontVariant(text='𞸂', kind=16), CharacterFontVariant(text='𞸢', kind=16), CharacterFontVariant(text='𞹂', kind=16), CharacterFontVariant(text='𞹢', kind=16), CharacterFontVariant(text='𞺂', kind=16), CharacterFontVariant(text='𞺢', kind=18)], 'د': [CharacterFontVariant(text='𞸃', kind=16), CharacterFontVariant(text='𞺃', kind=16), CharacterFontVariant(text='𞺣', kind=18)], 'و': [CharacterFontVariant(text='𞸅', kind=16), CharacterFontVariant(text='𞺅', kind=16), CharacterFontVariant(text='𞺥', kind=18)], 'ز': [CharacterFontVariant(text='𞸆', kind=16), CharacterFontVariant(text='𞺆', kind=16), CharacterFontVariant(text='𞺦', kind=18)], 'ح': [CharacterFontVariant(text='𞸇', kind=16), CharacterFontVariant(text='𞸧', kind=16), CharacterFontVariant(text='𞹇', kind=16), CharacterFontVariant(text='𞹧', kind=16), CharacterFontVariant(text='𞺇', kind=16), CharacterFontVariant(text='𞺧', kind=18)], 'ط': [CharacterFontVariant(text='𞸈', kind=16), CharacterFontVariant(text='𞹨', kind=16), CharacterFontVariant(text='𞺈', kind=16), CharacterFontVariant(text='𞺨', kind=18)], 'ي': [CharacterFontVariant(text='𞸉', kind=16), CharacterFontVariant(text='𞸩', kind=16), CharacterFontVariant(text='𞹉', kind=16), CharacterFontVariant(text='𞹩', kind=16), CharacterFontVariant(text='𞺉', kind=16), CharacterFontVariant(text='𞺩', kind=18)], 'ك': [CharacterFontVariant(text='𞸊', kind=16), CharacterFontVariant(text='𞸪', kind=16), CharacterFontVariant(text='𞹪', kind=16)], 'ل': [CharacterFontVariant(text='𞸋', kind=16), CharacterFontVariant(text='𞸫', kind=16), CharacterFontVariant(text='𞹋', kind=16), CharacterFontVariant(text='𞺋', kind=16), CharacterFontVariant(text='𞺫', kind=18)], 'م': [CharacterFontVariant(text='𞸌', kind=16), CharacterFontVariant(text='𞸬', kind=16), CharacterFontVariant(text='𞹬', kind=16), CharacterFontVariant(text='𞺌', kind=16), CharacterFontVariant(text='𞺬', kind=18)], 'ن': [CharacterFontVariant(text='𞸍', kind=16), CharacterFontVariant(text='𞸭', kind=16), CharacterFontVariant(text='𞹍', kind=16), CharacterFontVariant(text='𞹭', kind=16), CharacterFontVariant(text='𞺍', kind=16), CharacterFontVariant(text='𞺭', kind=18)], 'س': [CharacterFontVariant(text='𞸎', kind=16), CharacterFontVariant(text='𞸮', kind=16), CharacterFontVariant(text='𞹎', kind=16), CharacterFontVariant(text='𞹮', kind=16), CharacterFontVariant(text='𞺎', kind=16), CharacterFontVariant(text='𞺮', kind=18)], 'ع': [CharacterFontVariant(text='𞸏', kind=16), CharacterFontVariant(text='𞸯', kind=16), CharacterFontVariant(text='𞹏', kind=16), CharacterFontVariant(text='𞹯', kind=16), CharacterFontVariant(text='𞺏', kind=16), CharacterFontVariant(text='𞺯', kind=18)], 'ف': [CharacterFontVariant(text='𞸐', kind=16), CharacterFontVariant(text='𞸰', kind=16), CharacterFontVariant(text='𞹰', kind=16), CharacterFontVariant(text='𞺐', kind=16), CharacterFontVariant(text='𞺰', kind=18)], 'ص': [CharacterFontVariant(text='𞸑', kind=16), CharacterFontVariant(text='𞸱', kind=16), CharacterFontVariant(text='𞹑', kind=16), CharacterFontVariant(text='𞹱', kind=16), CharacterFontVariant(text='𞺑', kind=16), CharacterFontVariant(text='𞺱', kind=18)], 'ق': [CharacterFontVariant(text='𞸒', kind=16), CharacterFontVariant(text='𞸲', kind=16), CharacterFontVariant(text='𞹒', kind=16), CharacterFontVariant(text='𞹲', kind=16), CharacterFontVariant(text='𞺒', kind=16), CharacterFontVariant(text='𞺲', kind=18)], 'ر': [CharacterFontVariant(text='𞸓', kind=16), CharacterFontVariant(text='𞺓', kind=16), CharacterFontVariant(text='𞺳', kind=18)], 'ش': [CharacterFontVariant(text='𞸔', kind=16), CharacterFontVariant(text='𞸴', kind=16), CharacterFontVariant(text='𞹔', kind=16), CharacterFontVariant(text='𞹴', kind=16), CharacterFontVariant(text='𞺔', kind=16), CharacterFontVariant(text='𞺴', kind=18)], 'ت': [CharacterFontVariant(text='𞸕', kind=16), CharacterFontVariant(text='𞸵', kind=16), CharacterFontVariant(text='𞹵', kind=16), CharacterFontVariant(text='𞺕', kind=16), CharacterFontVariant(text='𞺵', kind=18)], 'ث': [CharacterFontVariant(text='𞸖', kind=16), CharacterFontVariant(text='𞸶', kind=16), CharacterFontVariant(text='𞹶', kind=16), CharacterFontVariant(text='𞺖', kind=16), CharacterFontVariant(text='𞺶', kind=18)], 'خ': [CharacterFontVariant(text='𞸗', kind=16), CharacterFontVariant(text='𞸷', kind=16), CharacterFontVariant(text='𞹗', kind=16), CharacterFontVariant(text='𞹷', kind=16), CharacterFontVariant(text='𞺗', kind=16), CharacterFontVariant(text='𞺷', kind=18)], 'ذ': [CharacterFontVariant(text='𞸘', kind=16), CharacterFontVariant(text='𞺘', kind=16), CharacterFontVariant(text='𞺸', kind=18)], 'ض': [CharacterFontVariant(text='𞸙', kind=16), CharacterFontVariant(text='𞸹', kind=16), CharacterFontVariant(text='𞹙', kind=16), CharacterFontVariant(text='𞹹', kind=16), CharacterFontVariant(text='𞺙', kind=16), CharacterFontVariant(text='𞺹', kind=18)], 'ظ': [CharacterFontVariant(text='𞸚', kind=16), CharacterFontVariant(text='𞹺', kind=16), CharacterFontVariant(text='𞺚', kind=16), CharacterFontVariant(text='𞺺', kind=18)], 'غ': [CharacterFontVariant(text='𞸛', kind=16), CharacterFontVariant(text='𞸻', kind=16), CharacterFontVariant(text='𞹛', kind=16), CharacterFontVariant(text='𞹻', kind=16), CharacterFontVariant(text='𞺛', kind=16), CharacterFontVariant(text='𞺻', kind=18)], 'ٮ': [CharacterFontVariant(text='𞸜', kind=16), CharacterFontVariant(text='𞹼', kind=16)], 'ں': [CharacterFontVariant(text='𞸝', kind=16), CharacterFontVariant(text='𞹝', kind=16)], 'ڡ': [CharacterFontVariant(text='𞸞', kind=16), CharacterFontVariant(text='𞹾', kind=16)], 'ٯ': [CharacterFontVariant(text='𞸟', kind=16), CharacterFontVariant(text='𞹟', kind=16)], 'ه': [CharacterFontVariant(text='𞸤', kind=16), CharacterFontVariant(text='𞹤', kind=16), CharacterFontVariant(text='𞺄', kind=16)]}
